% Copyright (C) 2014 Aravindh Krishnamoorthy
% aravindh.krishnamoorthy@ericsson.com, aravindh.k@ieee.org
% Released under the BSD license.
% Symmetric QR Algorithm with Permutations, arXiv:1402.5086.
%
% This script demostrates the Symmetric QR algorithm with Permutation, and 
% compares the performance to classic QR algorithm and variants.
N = 4 ;

NR = 10000 ;
NI = 50 ;
NA = 6 ;

% ANY_SYMMETRIC: 0 -> Positive definite only, 1 -> Any invertible symmetric
% matrix.
ANY_SYMMETRIC = 1 ;

randn('state', 100); %555) ;
err = zeros(NA, NI) ;
maxe = -Inf*ones(NA, NI) ;
mine = Inf*ones(NA, NI) ;
for nr=1:NR
    if mod(nr,10) == 0
        fprintf('%d ', nr) ;
    end
    
    while 1
        A = complex(randn(N,N), randn(N,N)) ;
        if ANY_SYMMETRIC
            A = (A+A')/2 ;
        else
            A = A'*A ;
        end
        if ~any(eig(A) == 0)
            break ;
        end
    end 
    Q = eye(N) ;
    E = sort(eig(A)) ;
    
    A1 = A ;
    A2 = A ;
    A3 = hess(A) ;
    A4 = hess(A) ;
    A5 = A ;
    A6 = A ;
    
    for ni=1:NI
        k = 1 ;
        % Algorithm 1
        leges{k} = 'BIC' ;
        A1 = onestep_bic(A1) ;
        D = sort(diag(A1)) ;
        te = norm(E-D)^2 ;
        err(k,ni) = err(k,ni) + te ;        
        if te < mine(k,ni); mine(k,ni) = te; end
        if te > maxe(k,ni); maxe(k,ni) = te; end

        k = k + 1 ;
        % Algorithm 2
        leges{k} = 'QR' ;
        A2 = onestep_qr(A2) ;
        D = sort(diag(A2)) ;
        te = norm(E-D)^2 ;
        err(k,ni) = err(k,ni) + te ;
        if te < mine(k,ni); mine(k,ni) = te; end
        if te > maxe(k,ni); maxe(k,ni) = te; end
        
        k = k + 1 ;
        % Algorithm 3
        leges{k} = 'QRH' ;
        A3 = onestep_qr(A3) ;
        D = sort(diag(A3)) ;
        te = norm(E-D)^2 ;
        err(k,ni) = err(k,ni) + te ;
        if te < mine(k,ni); mine(k,ni) = te; end
        if te > maxe(k,ni); maxe(k,ni) = te; end

        k = k + 1 ;
        % Algorithm 4
        leges{k} = 'QRS' ;
        A4 = onestep_qrs(A4) ;
        D = sort(diag(A4)) ;
        te = norm(E-D)^2 ;
        err(k,ni) = err(k,ni) + te ;        
        if te < mine(k,ni); mine(k,ni) = te; end
        if te > maxe(k,ni); maxe(k,ni) = te; end

        k = k + 1 ;
        % Algorithm 5
        leges{k} = 'DO' ;
        A5 = onestep_diord(A5) ;
        D = sort(diag(A5)) ;
        te = norm(E-D)^2 ;
        err(k,ni) = err(k,ni) + te ;        
        if te < mine(k,ni); mine(k,ni) = te; end
        if te > maxe(k,ni); maxe(k,ni) = te; end
      
        k = k + 1 ;
        % Algorithm 6
        leges{k} = 'CO' ;
        [A6,Q6] = onestep_co(A6) ;
        Q = Q*Q6 ;
        D = sort(diag(A6)) ;
        te = norm(E-D)^2 ;
        err(k,ni) = err(k,ni) + te ;
        if te < mine(k,ni); mine(k,ni) = te; end
        if te > maxe(k,ni); maxe(k,ni) = te; end
    end
end

fprintf ('\n') ;

% Normalise and transpose
err = err/NR ;
err = err' ;

figure() ;
semilogy(err(:,1), '--sk') ;
hold on ;
semilogy(err(:,2), '-ok') ;
semilogy(err(:,3), '-xk') ;
semilogy(err(:,4), '-*k') ;
semilogy(err(:,5), '-+k') ;
semilogy(err(:,6), '-dk') ;
grid on ;
title('Average of Errors') ;
legend(leges) ;
xlabel('Iterations \rightarrow') ;
ylabel('E_k^2 \rightarrow') ;

% figure() ;
% semilogy(maxe') ;
% grid on ;
% title('Max of Errors') ;
% legend(leges) ;

